/*******************************************************************************
  +-------------+
  | PUBBLI-STOP |
  +-------------+

  Apparecchio che rifiuta automaticamente le chiamate con numerazione anonima,
  come quelle pubblicitarie commerciali o statistiche.

--------------------------------------------------------------------------------
  Autore : SPU (stefano.purchiaroni@email.it)
  Data:    10.03.2014
--------------------------------------------------------------------------------

Descrizione del circuito:

  Lo squillo di chiamata viene rilevato da un fotoaccoppiatore che isola la MCU
  dalla linea telefonica. La ricezione dello squillo fa risvegliare la MCU dallo
  stato di basso consumo. Questa a sua volta risveglia il chip di ricezione
  MT88E39 che si mette in ascolo del flusso di dati modulato nel formato FSK.

  Il segnale FSK contenente i dati CLIP (Caller Line Identifier Presentation)
  viene livellato ed immesso nei pin di input del chip MT88E39, che lo demodula
  e lo ritrasmette tramite linea seriale a tre pin verso la MCU.

  La MCU  un PIC 12F683 con clock interno a 4 MHz. Il suo programma analizza il
  flusso dati in entrata. Se viene rilevata una chiamata con dati CLIP che
  confermano l'assenza della numerazione del chiamante, la MCU impegna la linea
  telefonica per due secondi, realizzando cos un riaggancio automatico.

  L'impegno della linea  ottenuto con un SSR (Solid State Relay) bidirezionale
  prodotto per usi telefonici. Il suo consumo  di 1.5-2 mA.

  L'alimentazione  fornita da un modulo Buck Converter (inverter step-down) che
  assorbe 1 mA dalla linea telefonica. Questo assorbimento  4 volte pi piccolo
  di quello che viene interpretato come "impegno linea" dalla centrale ma altri
  apparecchi che consumano corrente possono nell'insieme ottenere l'effetto di
  impegnare la linea. Verificare che non vi siano altri assorbimenti prima di
  installare l'apparecchio.

  Elementi di protezione (VDR) sono inseriti per prevenire danni derivanti da
  scariche elettriche che si propagano sulla linea telefonica.

--------------------------------------------------------------------------------

Tabella dei consumi di energia elettrica:

  +----------------------+-----------+------------+-------------------------+
  |STATO                 | CORRENTE  | DURATA     | COMPONENTI ATTIVE       |
  +----------------------+-----------+------------+-------------------------+
  |Consumo in Stand-By   | 0.075 mA  | indefinita |  MCU sby MT off SSR off |
  |Dopo il primo squillo | 2.6   mA  | 2s         |  MCU on  MT on  SSR off |
  |Se numerazione ok     | 0.8   mA  | 2..12s     |  MCU on  MT off SSR off |
  |Se numerazione assente| 2.5   mA  | 2s         |  MCU on  MT off SSR on  |
  +----------------------+-----------+------------+-------------------------+

--------------------------------------------------------------------------------
Note:

  * Richiede l'attivazione del servizio "Visualizzazione numero chiamante"
    sull'utenza, generalmente a pagamento da parte degli operatori.

  * L'apparecchio pu essere connesso a qualsiasi presa telefonica e non
    necessita di alimentazioni esterne.

--------------------------------------------------------------------------------
Registro delle modifiche:
-------------------+------------------------------------------------------------
 SPU 30.04.2014    | Creazione del programma e realizzazione del prototipo
*******************************************************************************/

//==============================================================================
// Costanti
//==============================================================================

// TIMEOUT : Tempo di attesa prima di tornare allo stato di basso consumo.
// Ogni squillo resetta il conteggio. Unit in centesimi di secondo.
unsigned int TimeOutTime   = 600;

// Tempo di impegno della linea in millesimi di secondo
#define EngageTime 2000

// Costanti generali
#define True       1
#define False      0
#define ON         1
#define OFF        0

//==============================================================================
// Definizione dei pin di I/O adattati al layout del PCB  (GP3 solo input!)
//==============================================================================

#define DAT        F2                // Input dal pin Data del MT88E39
#define DR         F3                // Input dal pin Data Ready del MT88E39
#define DCLK       F1                // Input dal pin Data Clock del MT88E39
#define RNG        F5                // Input dal circuito rivelatore di chiamata
#define LINE       F0                // Output al Rel e Led (opz) di impegno linea
#define PWDN       F4                // Output al Power Down del MT88E39

//==============================================================================
// Tipi e Variabili globali
//==============================================================================

// Definizione degli stati dell'automa di raccolta e gestione dei Byte ricevuti
enum STS {
  STS_STBY=0,                        // Automa inattivo
  STS_MTYP,                          // Acquisizione del Tipo di messaggio
  STS_MLEN,                          // Acquisizione della Lunghezza in byte
  STS_DATA,                          // Colleziona il byte nella stringa Data/Ora
  STS_NUMB,                          // Colleziona il byte nella stringa Numero
  STS_FTYP,                          // MDMF. Acquisizione del Tipo di Campo
  STS_FLEN,                          // MDMF. Acquisizione della lunghezza del campo (n. di byte)
  STS_FCHR,                          // MDMF. Colleziona il byte nel campo corrente
  STS_CKSM,                          // Colleziona la Checksum a analizza i dati ricevuti
  STS_FINE                           // Stato terminale. La gestione del CLIP  avvenuta.
};

// Variabili a livello modulo
unsigned short CurBit;               // Bit corrente, raccolto da GPIO
unsigned short CurByte;              // Byte corrente, raccolto un bit alla volta
unsigned short Idx;                  // Indice del byte corrente nel flusso
unsigned short MLen;                 // Lunghezza del messaggio corrente
unsigned short MTyp;                 // Tipo di messaggio
unsigned short FLen;                 // Lunghezza del campo corrente
unsigned short FTyp;                 // Tipo di campo
unsigned short Stato;                // Stato dell'automa
unsigned short DataSum;              // Somma di tutti i byte troncata a 8 bit
char *CurField;                      // Puntatore al campo corrente
char DateTime[8];                    // Stringa contenente data e ora
char Nome[30];                       // Stringa contenente il nome
char Number[20];                     // Stringa contenente il numero di telefono
unsigned short ChkSum;               // Checksum dichiarata dal messaggio
unsigned int CntTimer;               // Conteggio dei tic (scatti) del Timer
char *CurString;                     // Stringa in corso di acquisizione
unsigned int StringLen;              // Lunghezza attesa della stringa corrente
unsigned int StringIdx;              // Numero di caratteri acquisiti nella stringa corrente
unsigned short Cur_GPIO;             // Valore corrente dei pin di input
unsigned short Prev_GPIO;            // Valore precedente dei pin di input

//==============================================================================
// Funzioni usate dal Main
//==============================================================================

unsigned short EndData() {
//------------------------------------------------------------------------------
// Questa funzione torna True se  stata raggiunta la fine del messaggio
//------------------------------------------------------------------------------
unsigned char ret;

  if (Idx >= MLen)     ret = True;	 // Fine del messaggio
  else                 ret = False;
  return ret;
}

unsigned short GetCurByte() {
//------------------------------------------------------------------------------
// Questa funzione fornisce al chiamante il prossimo byte del dato raccolto durante la ricezione,
// ed aggiorna la somma dei valori letti, da usare per la verifica finale della Checksum.
//------------------------------------------------------------------------------
unsigned short b;

  b = CurByte;                       // Copia byte corrente
  CurByte = 0;                       // Reset byte corrente
  Idx++;					                   // Conteggo byte acquisiti
  DataSum += b;                      // Aggiorna la somma dei byte letti
  return b;   				               // Ritorna la copia del byte corrente
}

void InitString(char *s, int Len) {
//------------------------------------------------------------------------------
// Inizializza la stringa corrente predisponendone l'acquisizione byte a byte
//------------------------------------------------------------------------------

  CurString = s;                     // Punta alla stringa specificata
  CurString[0] = 0;                  // Inizializzazione della stringa di output
  StringIdx = 0;                     // Azzera il contatore di caratteri acquisiti
  StringLen = Len;                   // Iposta la lunghezza specificata
}

void CollectString() {
//------------------------------------------------------------------------------
// Questa funzione acquisisce un byte nella stringa corrente
//------------------------------------------------------------------------------
unsigned short c;

  c=GetCurByte();                    // Acquisizione del prossimo byte
  strncat(CurString,&c,1);           // Giustapposizione del carattere alla stringa
  StringIdx++;                       // Incrementa il contatore di caratteri aqcuisiti
}

unsigned short EndString() {
//------------------------------------------------------------------------------
// Torna True se la stringa corrente  stata completamente acquisita
//------------------------------------------------------------------------------
unsigned short ret;

  if (StringIdx<StringLen) ret = False;
  else                     ret = True;
  return ret;
}

void RifiutaChiamata() {
//------------------------------------------------------------------------------
// Impegno, pausa, e disimpegno delle linea telefonica.
//------------------------------------------------------------------------------
  GPIO.LINE=1;
  Delay_ms(EngageTime);
  GPIO.LINE=0;
}

void ProcessByte() {
//------------------------------------------------------------------------------
// Automa a stati finiti per la gestione del byte corrente
//------------------------------------------------------------------------------

  switch (Stato) {
    case STS_MTYP :                  // Acquisizione del Tipo di Messaggio
      MTyp = GetCurByte(); 	                 // Tipo di messaggio SDMF o MDMF
      if ((MTyp==0x80)||(MTyp==0x04)) {      // Ignora Seizure 0x55 o sua coda incompleta
        DataSum = MTyp;                      // Inizializza sommatoria dei byte con questo primo
        Idx = 1;                             // Inizializza contatore byte acquisiti
        Stato = STS_MLEN;                    // Prossimo stato: Lunghezza del messaggio
      }
      break;
    case STS_MLEN :                  // Acquisizione della Lunghezza in byte
      MLen = GetCurByte(); 	                 // Lunghezza (escluso byte di checksum)
      MLen++;                                // Collezioneremo anche la checksum nella sommatoria
      switch (MTyp) {
        case 0x04:                           // Caso SDMF (Single Data Message Format)
          InitString(DateTime,8);            // Prepara acquisizione Data/Ora
          Stato = STS_DATA;                  // Prossimo stato: SMDF-Data/Ora
          break;
        case 0x80:                           // Caso MDMF (Multiple Data Message Format)
          Stato = STS_FTYP;                  // Prossimo stato: MDMF-TipoCampo
          break;
        default:                             // (!) Messaggio di tipo non conosciuto
          GPIO.PWDN = 1;                     // Power-Down MT88E39
          Stato = STS_FINE;                  // Prossimo stato: Fine CLIP
          break;
      }
      break;
    case STS_DATA :                  // SMDF: Colleziona byte Data/Ora
      CollectString();                       // Acquisizione del byte nella stringa corrente
      if (EndString()==True) {
        InitString(Number,MLen-8);           // Prepara acquisizione Numero
        Stato = STS_NUMB;                    // Prossimo stato: SMDF-NumeroDiTelefono
      }
      break;
    case STS_NUMB :                  // SMDF: Colleziona byte Numero di telefono
      CollectString();                       // Acquisizione del byte nella stringa corrente
      if (EndString()==True) {
        Stato = STS_CKSM;                    // Prossimo stato: Checksum
      }
      break;
    case STS_FTYP :                  // MDMF: Acquisizione del Tipo di Campo
      FTyp = GetCurByte(); 		               // Tipo di campo
      Stato = STS_FLEN;                      // Prossimo stato: MDMF-LunghezzaCampo
      break;
    case STS_FLEN :                  // MDMF: Acquisizione della lunghezza del campo, in byte
      FLen = GetCurByte(); 		               // Lunghezza del campo
      switch (FTyp) {                        // Preparazione della stringa da acquisire
        case 0x01: InitString(DateTime,FLen); break;// Inizializza stringa Data e ora
        case 0x02: InitString(Number,FLen);   break;// Inizializza stringa Numero
        case 0x04: InitString(Number,FLen);   break;// Inizializza stringa Numero (assente)
        case 0x07: InitString(Nome,FLen);     break;// Inizializza stringa Nome
        case 0x08: InitString(Nome,FLen);     break;// Inizializza stringa Nome (assente)
      }
      Stato = STS_FCHR;                      // Prossimo stato: MDMF-CarattereStringa
      break;
    case STS_FCHR :                  // MDMF: Colleziona byte nel campo corrente fino a fine stringa
      CollectString();
      if (EndString()==True)
        if (EndData()==True) Stato = STS_CKSM;     // Prossimo stato: Checksum
        else                 Stato = STS_FTYP;     // Prossimo stato: MFDMF-TipoCampo altro campo
      break;
    case STS_CKSM :                  // Acquisizione della checksum e analisi dei dati ricevuti
      ChkSum = GetCurByte();	                     // Checksum
      GPIO.PWDN = 1;                               // Power-Down MT88E39
      Stato = STS_FINE;                            // Prossimo stato: Arresto automa
      if (DataSum==0) {                            // Checksum ok: Messaggio valido
        if ((Number[0]==0x50)||(Number[0]==0x4F)) {// Il numero inizia con P o O
          RifiutaChiamata();                       // OUT impulso sul Pin di impegno linea
        }
      }
      break;
  }
  CurByte=0;                         // Azzeramento del byte ormai utilizzato
}

//==============================================================================
// Funzioni della procedura di gestione degli Interrupt
//==============================================================================

void StartTimer() {
//------------------------------------------------------------------------------
// Impostazione del prescaler del Timer0, sua abilitazione e avvio
// Reset del contatore associato al timer. Uno scatto ogni 0.01 s
//------------------------------------------------------------------------------
  INTCON.T0IE = 1;                   // Abilita l'interrut sull'overflow del timer TMRO
  INTCON.T0IF = 0;                   // Reset Timer0 Interrupt Flag
  TMR0        = 100;                 // Imposta valore iniziale di TMR0
}

void ResetTimer() {
//------------------------------------------------------------------------------
// Azzera CntTimer (bloccando l'interrupt durrante l'impostazione)
//------------------------------------------------------------------------------
  INTCON.T0IE = 0;                   // Disabilita l'interrut dal timer TMRO
  CntTimer = 0;                      // Inizializza conteggio tic
  INTCON.T0IE = 1;                   // Abilita l'interrut dal timer TMRO
}

void StopTimer() {
//------------------------------------------------------------------------------
// Azzera CntTimer (bloccando l'interrupt durrante l'impostazione)
//------------------------------------------------------------------------------
  INTCON.T0IE = 0;                   // Disabilita l'interrut dal timer TMRO
  INTCON.T0IF = 0;                   // Reset Timer0 Interrupt Flag
}

void interrupt() {
//------------------------------------------------------------------------------
// Gestione degli interrupt. Interrogazione dei pin di ingresso e chiamata
// alle procedure di acquisizione e processamento del dato in arrivo.
//------------------------------------------------------------------------------

// Interrupt dovuto ad un cambiamento di un pin i/o abilitato
  if ((INTCON.GPIF==1) && (INTCON.GPIE==1)) {
    // Salvataggio GPIO precedente e corrente per analisi variazioni
    Prev_GPIO = Cur_GPIO;
    Cur_GPIO = GPIO;

    // Gestione del primo Ring Tone per l'avvio dell'automa
    if ((Prev_GPIO.RNG==0) && (Cur_GPIO.RNG==1))   {   // RNG 0-->1
      StartTimer();                  // Avvio (o Riavvio) del Timer
      ResetTimer();                  // Azzera conteggio tic del Timer
      if (Stato==STS_STBY) {         // Avvio dell'automa al primo Ring
        GPIO.PWDN = 0;               // Risveglio del MT88E39 per ricezione CLI
        Stato = STS_MTYP;            // Prossimo stato automa: Acquisizione Tipo messaggio
        CurByte = 0;                 // Azzera il byte corrente
      }
    }

    // Raccolta del bit DATA su transizione DCLK 0-->1
    // Nel caso peggiore abbiamo solo 6us per raccogliere il bit!
    if ((Prev_GPIO.DCLK==0) && (Cur_GPIO.DCLK==1)) {   // DCLK 0-->1
      CurByte >>= 1;                 // Sposta i bit presenti a destra
      CurByte.F7 = Cur_GPIO.DAT;     // Inserimento nuovo bit come MSB
    }

    // Analisi del Byte costruito su transizione DR 1-->0
    if ((Prev_GPIO.DR==1) && (Cur_GPIO.DR==0)) {       // DR 1-->0
      ProcessByte();                 // Automa di gestione del byte raccolto
    }
    
    INTCON.GPIF = 0;                 // Reset GP Interrupt Flag
  }

// Interrupt dovuto all'overflow del Timer: un tic trascorso.
  if ((INTCON.T0IF==1) && (INTCON.T0IE==1)) {
    CntTimer++;                      // Conta tic (0.01 s trascorsi)
    if (CntTimer>TimeOutTime) {      // Timeout per assenza Ring
      Stato = STS_STBY;              // Prossimo stato automa: ritorno a basso consumo
      GPIO.PWDN = 1;                 // Power-Down MT88E39
      StopTimer();                   // Arresto Timer
    }
    else
      StartTimer();                  // Riavvia TMR0 per prossimo tic
  }
}

void main() {
//==============================================================================
// MAIN : Impostazioni iniziali
//==============================================================================

  // L'avvio avviene con 64 ms di ritardo rispetto al Power-On (PWRTE abilitato)
  
  // Impostazioni hardware generali
  ANSEL         = 0;                 // Modo di impiego dei pin come Digital i/o
  CMCON0        = 0x07;              // CIN e COUT come I/O, Comparator off.
  OPTION_REG    = 0x85;              // Disabilita Pull-up e imposta il Prescaler del Timer0

  // Direzione dei pin di I/O (1: Input, 0: Output)
  TRISIO.DAT    = 1;                 // Input dal pin Data del MT88E39
  TRISIO.DR     = 1;                 // Input dal pin Data Ready del MT88E39
  TRISIO.DCLK   = 1;                 // Input dal pin Data Clock del MT88E39
  TRISIO.RNG    = 1;                 // Input dal circuito rivelatore di chiamata
  TRISIO.LINE   = 0;                 // Output al Rel e Led (opz) di impegno linea
  TRISIO.PWDN   = 0;                 // Output al Power Down del MT88E39

  // Abilitazione dei PIN sui quali cambiamenti va gestito l' Interrupt
  IOC.DAT       = 0;                 // Nessun interrupt impostato  su DAT
  IOC.DR        = 1;                 // Abilita interrupt-on-change su DR
  IOC.DCLK      = 1;                 // Abilita interrupt-on-change su DCLK
  IOC.RNG       = 1;                 // Abilita interrupt-on-change su RNG

  // Impostazioni iniziali dei pin di output
  GPIO.PWDN     = 1;                 // PWDN se 1 induce lo shutdown del MT88E39
  GPIO.LINE     = 0;                 // LINE se 1 si impegna la linea telefonica

  // Attesa stabilizzazione linea all'accensione per ignorare ronzio iniziale
  Delay_ms(4000);

  // Abilitazione generale interrupts
  INTCON.GIE    = 1;                 // Abilita Global unmasked Interrupt
  INTCON.GPIE   = 1;                 // Abilita Interrupt-On-Change
  INTCON.INTE   = 0;                 // Disabilita GP2/INT External Interrupt
  INTCON.INTF   = 0;                 // Reset GP2/INT external interrupt status

  // Inizializzazione dei valori correnti e precedenti dei pin di input
  Cur_GPIO      = 0;                 // Valore corrente dei pin
  Prev_GPIO     = 0;                 // Valore precedente dei pin

  // Inizializzazione dell'Automa
  Stato         = STS_STBY;          // Automa in attesa del primo squillo

//==============================================================================
// MAIN : Ciclo in stato di basso consumo. Operazioni gestite solo da interrupt
//==============================================================================

  while (True) {
    if (Stato==STS_STBY) {asm{sleep};} // MCU in Basso Consumo      ( 0.075 mA )
    else Delay_us(1);                  // MCU in Modalit operativa ( 0.8   mA )
  }
}//~!

